import numpy as np
import matplotlib.pyplot as plt
from typing import List, Dict
from sklearn.preprocessing import StandardScaler
from numpy.linalg import solve, cholesky
from scipy.spatial.distance import cdist
from scipy.optimize import minimize
from scipy.special import gamma, kv
import requests
import io
import copy


EPSILON = 1e-8

# Define Hyperparameter Class
class Hyperparam:
    def __init__(self, value = 1.0, bounds = [-np.inf, np.inf]):
        self.value: float = value
        self.bounds : list = bounds


# Define Kernel Base Class
class Kernel():
    def __init__(self, hyperparams):
        self.hyperparams = dict()
        self.initialize_hyperparams(hyperparams)

    def get_hyperparam_names(self):
        return sorted(self.hyperparams.keys())

    def initialize_hyperparams(self, hyperparams: Dict[str, Hyperparam]):
        # Initialize the default hyperparam values and bounds
        defaults = {
            'alpha' : Hyperparam(value = 1.0, bounds = [EPSILON, 5.0]),
            'lengthscale' : Hyperparam(value = 10, bounds = [EPSILON, 50.0]),
            'sigma_f' : Hyperparam(value = 0.7, bounds = [EPSILON, 2.0]),
            'nu' : Hyperparam(value = 30, bounds = [EPSILON, 50.0])
        }

        # Fill in the hyperparams with defaults (if necessary)
        for hp_name in hyperparams.keys():
            hp = hyperparams[hp_name]
            self.hyperparams[hp_name] = Hyperparam()
            if hp:
                self.hyperparams[hp_name].value = defaults[hp_name].value if \
                                                  hp.value is None else hp.value
                self.hyperparams[hp_name].bounds = defaults[hp_name].bounds if \
                                                   hp.bounds is None else hp.bounds
            else:
                self.hyperparams[hp_name] = copy.deepcopy(defaults[hp_name])

# Define the SE Kernel
class SquaredExponential(Kernel):
    def __init__(self, sigma_f: Hyperparam = None, lengthscale: Hyperparam = None):
        _hyperparams = {
            'sigma_f' : sigma_f,
            'lengthscale' : lengthscale,
        }
        super().__init__(_hyperparams)

    # Covariance function definition
    def cov(self, xi, xj):
        sf, l = self.hyperparams['sigma_f'], self.hyperparams['lengthscale']
        dist = cdist(xi, xj, 'sqeuclidean') 
        covariance = sf**2 * np.exp(-dist/(2*l**2))
        return covariance

# Define the RQ Kernel
class RationalQuadratic(Kernel):
    def __init__(self, sigma_f: Hyperparam = None, 
                lengthscale: Hyperparam = None, alpha: Hyperparam = None):
        _hyperparams = {
            'sigma_f' : sigma_f,
            'lengthscale' : lengthscale,
            'alpha' : alpha
        }
        super().__init__(_hyperparams)
    
    # Covariance function definition
    def cov(self, xi, xj):
        sf, l, a = self.hyperparams['sigma_f'], self.hyperparams['lengthscale'], \
                   self.hyperparams['alpha']
        dist = cdist(xi, xj, 'sqeuclidean')
        covariance = sf**2 * (1 + dist/(2*a * l**2))**(-a)
        return covariance

# Define the Matern Kernel
class Matern(Kernel):
    def __init__(self, nu: Hyperparam = None, lengthscale: Hyperparam = None):
        _hyperparams = {
            'nu' : nu,
            'lengthscale' : lengthscale,
        }
        super().__init__(_hyperparams)

    # Covariance function definition
    def cov(self, xi, xj):
        v, l = self.hyperparams['nu'], self.hyperparams['lengthscale']
        coeff = (2 ** (1 - v)) / gamma(v)
        dist = cdist(xi, xj, 'euclidean')
        dist[dist == 0] = EPSILON
        dist *= (np.sqrt(2 * v) / l)
        cov = coeff * dist**v * kv(v, dist)
        return cov


# Define Gaussian Process Class
class GaussianProcess():
    def __init__(self, kernel: Kernel):
        self.kernel = kernel
        self.sigma_n = Hyperparam(value=0.0, bounds=[EPSILON, 3])
        # Initialize data struct to store train/test points
        self.data = {
            'X' : np.array([]),
            'f' : np.array([]),
            'X*' : np.array([])
        }

    # Notational wrapper for the Kernel
    def _K(self, XiXj: str):
        Xi, Xj = str.split(XiXj,', ')
        cov = self.kernel.cov(self.data[Xi], self.data[Xj])
        return cov

    # Calculate conditional posterior (Rasmussen)
    # NOTE: there are redundant computations (strictly for readability)
    def predict(self):
        K, f, n = self._K, self.data['f'], len(self.data['X'])
        L = cholesky(K('X, X') + np.eye(n)*self.sigma_n.value**2)
        alpha = solve(L.T, solve(L, f))
        mu_post = K('X, X*').T @ alpha
        v = solve(L, K('X, X*'))
        cov_post = K('X*, X*') - (v.T @ v)
        return (mu_post, cov_post)
    
    # Negative Log-Likelihood for Kernel Hyperparameters
    def _nll_kernel_hyperparams(self, hyperparam_values):
        K, f, n = self._K, self.data['f'], len(self.data['X'])
        # Set the kernel parameters to the input from the optimizer
        for name, value in zip(self.kernel.get_hyperparam_names(), hyperparam_values):
            self.kernel.hyperparams[name] = value
        self.sigma_n.value = hyperparam_values[-1]
        
        # If the proposed optimizer values cause the matrix to be non-PSD, return inf
        try: L = cholesky(K('X, X') + np.eye(n)*self.sigma_n.value**2) 
        except: return np.inf
        
        alpha = solve(L.T, solve(L, f))
        loglik = -0.5*(f.T @ alpha) - np.trace(L)
        return -loglik
    
    def optimize_kernel_hyperparams(self):
        # Initialize the starting values for the hyperparams
        starting_vals = [self.kernel.hyperparams[name].value \
                         for name in self.kernel.get_hyperparam_names()]
        starting_vals.append(self.sigma_n.value) # append the sigma_n term
        
        # Set the bounds for the hyperparams
        hp_bounds = [self.kernel.hyperparams[name].bounds \
                     for name in self.kernel.get_hyperparam_names()]
        hp_bounds.append(self.sigma_n.bounds)
        
        # Minimize the marginal likelihood
        cost = minimize(self._nll_kernel_hyperparams, starting_vals, 
                        bounds=hp_bounds, method='L-BFGS-B')
        nll = cost.fun
        print('Optimized values for kernel hyperparameters (NLL = ' + str(nll) + ') :\n', \
              self.kernel.hyperparams, end=" ")
        print(', sigma_n: ', self.sigma_n.value)
        return nll

    def simulate_posterior_draws(self, mu_post, cov_post, num_draws):
        L_post = cholesky(cov_post + np.eye(cov_post.shape[1])*EPSILON)
        sim_draws = mu_post + L_post @ np.random.normal(size=(L_post.shape[1], num_draws))
        return sim_draws


response = requests.get('https://www.drolet.io/projects/e_data.npy')
# Load trajectories, Shape: [N (num trajectories) X 100 (num samples) X 2 (num dimensions)]
e_data = np.load(io.BytesIO(response.content))

# Extract a random trajectory for testing
test_idx = np.random.choice(len(e_data))
print('Using test idx:', test_idx)
test_traj = e_data[test_idx]

# Use non-test data solely for fitting the scalers
scale_data = np.delete(e_data, [test_idx], axis=0)
x_scale, y_scale = StandardScaler().fit(scale_data[:,:,0]), StandardScaler().fit(scale_data[:,:,1])

num_train_points = 10  # Number of points to condition on (training points)
domain_ends = (0, 100)

# Define domain and randomly select points on the domain for the test trajectory
domain = np.linspace(domain_ends[0], domain_ends[1], e_data.shape[1])
domain_idxs = sorted(np.random.choice(range(e_data.shape[1]), num_train_points, replace=False))
T_obs = domain[domain_idxs]

# Get corresponding X trajectory values
obs, obs_full = dict(), dict()
obs_full['X_dim'] = x_scale.transform(e_data[test_idx,:,0].reshape(1,-1)).flat
obs['X_dim'] = obs_full['X_dim'][domain_idxs]

# Get corresponding Y trajectory values
obs_full['Y_dim'] = y_scale.transform(e_data[test_idx,:,1].reshape(1,-1)).flat
obs['Y_dim'] = obs_full['Y_dim'][domain_idxs]

# Define the lengthscale default to 1/10 of the domain length
default_lengthscale = Hyperparam(value=0.1*len(T_obs), bounds=[EPSILON, len(domain)/2])

##### Create the GaussianProcess Instances #####
GPs = {
    'X_dim' : {
        'SE' : GaussianProcess(kernel=SquaredExponential(lengthscale = default_lengthscale)),
        'RQ' : GaussianProcess(kernel=RationalQuadratic(lengthscale = default_lengthscale)),
        'Matern' : GaussianProcess(kernel=Matern(lengthscale = default_lengthscale))
    },
    'Y_dim' : {
        'SE' : GaussianProcess(kernel=SquaredExponential(lengthscale = default_lengthscale)),
        'RQ' : GaussianProcess(kernel=RationalQuadratic(lengthscale = default_lengthscale)),
        'Matern' : GaussianProcess(kernel=Matern(lengthscale = default_lengthscale))
    }
}

##### Perform Conditional Inference on data dimensions with the different kernels #####
for dim in GPs.keys():
    gp_dim = GPs[dim]
    for kernel in gp_dim.keys():
        gp = gp_dim[kernel]
        gp.data['X'], gp.data['f'] = T_obs.reshape(-1,1), obs[dim].reshape(-1,1)
        gp.nll = gp.optimize_kernel_hyperparams()
        gp.data['X*'] = domain.reshape(-1,1)
        gp.mu_post, gp.cov_post = gp.predict()
        gp.sim_draws = gp.simulate_posterior_draws(gp.mu_post, gp.cov_post, num_draws=1000)

##### Plot the Conditional Posterior #####
n_std = 3
fig, axs = plt.subplots(nrows=2, ncols=3, figsize=(16, 10))

# Iterate over all GPs and find the best ones
best_kernels = { 'X_dim' : None, 'Y_dim' : None }
for i, dim in enumerate(GPs.keys()):
    gp_dim = GPs[dim]
    best_nll = np.inf
    for j, kernel in enumerate(gp_dim.keys()):
        # Find best GP for given dim
        gp = gp_dim[kernel]
        if gp.nll < best_nll:
            best_kernels[dim] = kernel
            best_nll = gp.nll

        # Plot ground truth
        ax = axs[i,j]
        ax.plot(domain, obs_full[dim], 'k--', label='Ground Truth')
        # Plot mean +/- N Std Devs
        std = np.std(gp.sim_draws, axis=1)
        ax.fill_between(domain, gp.mu_post.flat-n_std*std, 
                        gp.mu_post.flat+n_std*std, color='red', alpha=0.15, 
                        label='$\mu \pm$' + str(n_std) + ' $\sigma$')
        # Plot the training points
        ax.plot(T_obs, obs[dim], 'ko', linewidth=2, label='Train Points')
        # Plot the mean of the posterior
        ax.plot(domain, gp.mu_post, 'r-', lw=2, label='$\mu$')
        
        # Add plot descriptions
        ax.set_xlabel('$t$', fontsize=14)
        ax.set_ylabel(dim, fontsize=14)
        ax.set_title('Posterior draws for ' + dim + ' with ' + kernel + 
                     ', NLL: ' + str(round(gp.nll, 3)), fontsize=14)
        ax.legend(fontsize=9)
plt.show()

##### Plot the Predicted Letter #####
# Get the "best" GPs according to NLL values
gp_x, gp_y = GPs['X_dim'][best_kernels['X_dim']], GPs['Y_dim'][best_kernels['Y_dim']]
fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(10, 6))
# Plot original trajectory
x_data, y_data = e_data[test_idx,:,0].flat, e_data[test_idx,:,1].flat
ax.plot(x_data, y_data, 'k--', label='Ground Truth', lw=6)

# Plot the mean predicted trajectory
mu_x = x_scale.inverse_transform(gp_x.mu_post.reshape(1,-1)).flat
mu_y = y_scale.inverse_transform(gp_y.mu_post.reshape(1,-1)).flat
ax.plot(mu_x, mu_y, color='r', label='Predicted', lw=6, alpha=0.7)
ax.plot(x_data[domain_idxs], y_data[domain_idxs], 'bo', label='Train Points', markersize=12)
ax.legend(fontsize=12)
plt.show()